#ifndef BBF4C6D8_A133_C8FC_1285_ADC240274FF5
#define BBF4C6D8_A133_C8FC_1285_ADC240274FF5

#define YapicDI_Err_GotProvideForValue "Cannot provide injectables for simple value."
#define YapicDI_Err_GotInvalidStrategy "The 'strategy' param must be 'int' or 'callable' got: %R."
#define YapicDI_Err_GotInvalidStrategyInt "The 'strategy' param value is not a valid strategy constant: %R."
#define YapicDI_Err_InjectableNotFound "Not found suitable value for: %R."
#define YapicDI_Err_CallableArgument "Argument must be callable."
#define YapicDI_Err_CallableArgumentUntyped "Argument must have a type or a default value."
#define YapicDI_Err_CallableMustHaveKwOnly "Callable must have kwonly arguments."
#define YapicDI_Err_KwOnlyGetterMustHaveNameKw "Keyword argument resolver function muts have 'name' keyword only argument."
#define YapicDI_Err_ProvideArgMustIterable "The 'provide' argument must be iterable list or tuple."
#define YapicDI_Err_OneInjectorArg "Bad argument, must call with 'Injector' instance."
#define YapicDI_Err_RecursionError "Recusrion fund while injecting:\n%R"

#endif /* BBF4C6D8_A133_C8FC_1285_ADC240274FF5 */
